
#include "lightup_creator.h"

bool Solve::ProcessLoop(Stage &stage)
{
	for( ; ; ){
		bool updated = false;
		bool error = false;
		
		stage.Update(&updated, &error);
		++loopCount;
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(!updated){
			return true;
		}
	}
}

bool Solve::ProcessSubloop(Stage &stage, int count)
{
	for(int i = 0; i < count; i++){
		bool updated = false;
		bool error = false;
		
		stage.Update(&updated, &error);
		
		if(error){
			return false;
		}else if(!updated){
			return true;
		}
	}
	return true;
}

bool Solve::TryOnce(Stage &stage, bool find_cross, int count)
{
	Stage substage1(stage);
	Stage substage2(stage);
	int pos = 0;
	int try_pos[2];
	int try_type;
	
	for( ; ; ){
		pos = stage.GetTryPos2(pos, try_pos, &try_type);
		if(pos == -1){
			return false;
		}
		
		substage1.Reset(stage);
		substage1.SetLight(try_pos[0]);
		bool result1 = ProcessSubloop(substage1, count);
		
		substage2.Reset(stage);
		substage2.SetLight(try_pos[1]);
		bool result2 = ProcessSubloop(substage2, count);
		
		if(result1 && !result2){
			stage.SetLight(try_pos[0]);
			DumpStage(stage);
			++tryCount2[try_type + 2];
			return true;
		}else if(!result1 && result2){
			stage.SetLight(try_pos[1]);
			DumpStage(stage);
			++tryCount2[try_type + 2];
			return true;
		}else if(find_cross && result1 && result2){
			bool updated = false;
			stage.FindCross(substage1, substage2, &updated);
			if(updated){
				DumpStage(stage);
				++tryCount2[try_type + 2];
				return true;
			}
		}
		
		++tryCount2[try_type];
	}
}

bool Solve::TryTwice(Stage &stage, int count)
{
	Stage substage1(stage);
	Stage substage2(stage);
	Stage substage3(stage);
	int pos = 0;
	int try_pos[3];
	bool light;
	
	for( ; ; ){
		pos = stage.GetTryPos3(pos, try_pos, &light);
		if(pos == -1){
			return false;
		}
		
		++tryCount2[4];
		
		substage1.Reset(stage);
		substage2.Reset(stage);
		substage3.Reset(stage);
		
		if(light){
			substage1.SetLight(try_pos[0]);
			substage2.SetLight(try_pos[1]);
			substage3.SetLight(try_pos[2]);
		}else{
			substage1.SetCross(try_pos[0]);
			substage2.SetCross(try_pos[1]);
			substage3.SetCross(try_pos[2]);
		}
		
		bool result1 = ProcessSubloop(substage1, count);
		bool result2 = ProcessSubloop(substage2, count);
		bool result3 = ProcessSubloop(substage3, count);
		
		if(result1 && !result2 && !result3){
			stage.SetLight(try_pos[0]);
			DumpStage(stage);
			return true;
		}else if(!result1 && result2 && !result3){
			stage.SetLight(try_pos[1]);
			DumpStage(stage);
			return true;
		}else if(!result1 && !result2 && result3){
			stage.SetLight(try_pos[2]);
			DumpStage(stage);
			return true;
		}
	}
}

bool Solve::Process(Stage &stage, bool dump)
{
	DumpStage(stage);
	
	for( ; ; ){
		if(!ProcessLoop(stage)){
			if(dump){
				DumpResult(stage);
			}
			return false;
		}
		if(stage.IsSolved()){
			if(dump){
				DumpResult(stage);
			}
			return true;
		}
		
		if(TryOnce(stage, false, 1)){
			++tryCount[0];
		}else if(TryOnce(stage, true, 1)){
			++tryCount[1];
		}else if(TryOnce(stage, false, 2)){
			++tryCount[1];
		}else if(TryOnce(stage, false, 10)){
			++tryCount[2];
		}else if(TryTwice(stage, 1)){
			++tryCount[3];
		}else if(TryTwice(stage, 2)){
			++tryCount[4];
		}else if(TryTwice(stage, 10)){
			++tryCount[5];
		}else{
			if(dump){
				DumpResult(stage);
			}
			return false;
		}
	}
}

bool Solve::ProcessFast(Stage &stage)
{
	for( ; ; ){
		if(!ProcessLoop(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		if(TryOnce(stage, true, 1)){
		}else if(TryOnce(stage, false, 10)){
		}else if(TryTwice(stage, 10)){
		}else{
			return false;
		}
	}
}

bool Solve::ProcessSimple(Stage &stage)
{
	for( ; ; ){
		if(!ProcessLoop(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		if(TryOnce(stage, true, 2)){
		}else{
			return false;
		}
	}
}

char Solve::GetDifficulty()
{
	static const double coef[] = {
		0.631687, -1.054039, -0.471749, -2.992082, -0.002099, 0.0178, 
		0.531462, 0, -0.310276
	};
	static const double thre[] = {
		10, 15, 20, 25, 60,
	};
	
	double d = loopCount * coef[0] + tryCount[0] * coef[1] + 
		tryCount[1] * coef[2] + tryCount[2] * coef[3] + 
		tryCount2[0] * coef[4] + tryCount2[1] * coef[5] + 
		tryCount2[2] * coef[6] + tryCount2[3] * coef[7] + coef[8];
	
	for(char i = 0; i < 5; i++){
		if(d < thre[i]){
			return i + 1;
		}
	}
	return 5;
}

bool SolveProblem(StageInfo &info)
{
	Stage stage(info);
	Solve solve;
	
	bool result = solve.Process(stage, true);
	
//	int loopCount = solve.GetLoopCount();
//	int *tryCount = solve.GetTryCount();
//	int *tryCount2 = solve.GetTryCount2();
	char difficulty = solve.GetDifficulty();
	
	if(result){
		printf("finished. difficulty = %d\n", difficulty);
//		printf("%d, %d, %d, %d, %d, %d, %d, %d\n", 
//			loopCount, tryCount[0], tryCount[1], tryCount[2], tryCount2[0], 
//			tryCount2[1], tryCount2[2], tryCount2[3]);
		return true;
	}else{
		printf("failed. difficulty = %d\n", difficulty);
		return false;
	}
}
